/* exportingDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "exportingDialog.h"
#include "song.h"
#include "playlistProxy.h"
#include "get_filename_extension.h"
#include "get_basename.h"
#include "queryResult.h"
#include "database.h"
#include "expand_variables.h"
#include "infoMessage.h"
#include "iso8859_to_ascii.h"
#include "mkdir_p.h"
#include "artist.h"
#include "song.h"
#include "cd.h"
#include "volume.h"
#include "guess_file_type.h"
#include "protect_from_shell.h"

#include <unistd.h>
#include <time.h>

ExportingDialog::
ExportingDialog(const Dataset &data, 
		const PlaylistProxy &playlist, int start_row,
		bool symlink, int convert_type,
		const string &root_dir,
		const string &filename_pattern,
		double max_time) :
  PalomaDialog(data),
  _symlink(symlink),
  _convert_type(convert_type),
  _root_dir(root_dir),
  _filename_pattern(filename_pattern),
  _max_time(max_time),
  _cancel_button("Cancel")
{
  _canceled = false;
  //  _title.set_text("Exporting song files");

  // Build up a list in _entries of all the songs to be read.
  _total_length = 0.0;
  int row = start_row;
  int num_rows = playlist.get_num_rows();
  assert(num_rows > start_row);

  if (playlist.get_primary_table() == TI_songs) {
    // If the playlist is based on songs, it's simple.

    TableDataProxy song(TI_songs, playlist.get_row(row)->_primary_key);
    while (row < num_rows && 
	   (row == start_row || _max_time < 0 ||
	    _total_length + song[SFI_length].get_float() < _max_time * 60.0)) {
      Entry e;
      e._song_id = song.get_primary_key();
      e._cd_id = playlist.get_row(row)->_cd_id;
      _entries.push_back(e);

      _total_length += song[SFI_length].get_float();
      
      row++;
      if (row < num_rows) {
	song = TableDataProxy(TI_songs, playlist.get_row(row)->_primary_key);
      }
    }

  } else {
    // If it's based on CD's, it just requires an extra step.

    bool stop = false;
    while (row < num_rows && !stop) {
      int cd_id = playlist.get_row(row)->_primary_key;
      QueryResult qr = music_db->sql_command
	("SELECT cd_tracks.song_id, songs.length FROM songs, cd_tracks"
	 " WHERE cd_tracks.song_id=songs.song_id"
	 " AND cd_tracks.cd_id=" + DataCell(cd_id) +
	 " ORDER BY track_num");
      int num_tracks = qr.get_num_rows();
      for (int t = 0; t < num_tracks && !stop; t++) {
	int song_id = qr.get_cell(t, 0).get_int();
	float length = qr.get_cell(t, 1).get_float();
	if (_max_time >= 0 && _total_length + length >= _max_time * 60.0) {
	  stop = true;
	} else {
	  _total_length += length;
	  Entry e;
	  e._song_id = song_id;
	  e._cd_id = cd_id;
	  _entries.push_back(e);
	}
      }
      
      row++;
    }
  }

  _current_length = 0.0;
  _current_index = 0;
  _progress_bar.configure(0.0, 0.0, _total_length);
  _progress_bar.set_show_text(true);
  _last_return = time(NULL);

  layout_window();
  setup();

  connect_to_method(_shell.success, this, &export_success);
  connect_to_method(_shell.failure, this, &export_failure);

  if (!export_current()) {
    // We'll set a timeout for a quarter second longer just to give
    // the window a chance to pop up before we destroy it.  (Surely
    // there's a better way!)
    connect_to_method(Gtk_Main::timeout(250), this, &delete_timeout);
  }
}

ExportingDialog::
~ExportingDialog() {
}

void ExportingDialog::
export_success() {
  if (_canceled || !advance()) {
    // If that was the last track, close the window.
    delete_self();
  }
}

void ExportingDialog::
export_failure() {
  if (!_canceled) {
    new InfoMessage("Failure while exporting.");
  }
  delete_self();
}

void ExportingDialog::
cancel_button_clicked() {
  _canceled = true;
  _shell.kill();
}

gint ExportingDialog::
delete_timeout() {
  delete_self();
  return false;
}

int ExportingDialog::
back_from_idle() {
  // This function is only invoked if we're doing symlinks, which are
  // very quick, and don't involve waiting around for a command to
  // complete.  From time to time we do need to yield control, so we
  // hang this hook on idle() and exit.  When we come back, we want to
  // pick up with what we were doing.
  export_success();
  return false;
}


void ExportingDialog::
layout_window() {
  set_default_size(200, 0);
  set_title("Exporting song files");

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  /*
  _box1.pack_start(_title, true, false, 0);
  _title.show();
  */
  _box1.pack_start(_progress_bar, true, false, 0);
  _progress_bar.show();
  _box1.pack_start(_song_name, true, false, 0);
  _song_name.show();

  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();

  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
};


bool ExportingDialog::
advance() {
  TableDataProxy done_song(TI_songs, _entries[_current_index]._song_id);
  _current_length += done_song[SFI_length].get_float();

  _current_index++;

  return export_current();
}

bool ExportingDialog::
export_current() {
  if (_current_index >= _entries.size()) {
    // All done.
    _current_length = _total_length;
    _progress_bar.set_value(_current_length);
    _song_name.set_text(string());
    return false;
  }

  TableDataProxy song(TI_songs, _entries[_current_index]._song_id);
  TableDataProxy cd(TI_cds, _entries[_current_index]._cd_id);
  TableDataProxy performer = song[SFI_performer].get_foreign(TI_artists);
  TableDataProxy writer = song[SFI_writer].get_foreign(TI_artists);
  TableDataProxy cdperformer = cd[CFI_performer].get_foreign(TI_artists);
  TableDataProxy cdwriter = cd[CFI_writer].get_foreign(TI_artists);
  int track_num = 0;

  if (cd.get_primary_key() < 0) {
    // If the song doesn't know its CD, try to look it up.  This is
    // ambiguous, because a song might appear on multiple CD's.  But
    // that's what you get if you make a query without specifying the
    // CD explicitly.
    QueryResult qr = music_db->sql_command
      ("SELECT cd_id, track_num FROM cd_tracks"
       " WHERE song_id=" + DataCell(song.get_primary_key()) +
       " ORDER BY cd_id");
    if (qr.get_num_rows() > 0) {
      cd = qr.get_cell(0, 0).get_foreign(TI_cds);
      track_num = qr.get_cell(0, 1).get_int();
    }

  } else {
    // Otherwise, if the song did know its CD, we still need to get
    // its track number.
    QueryResult qr = music_db->sql_command
      ("SELECT track_num FROM cd_tracks"
       " WHERE song_id=" + DataCell(song.get_primary_key()) +
       " AND cd_id=" + DataCell(cd.get_primary_key()) +
       " ORDER BY track_num");
    if (qr.get_num_rows() > 0) {
      track_num = qr.get_cell(0, 0).get_int();
    }
  }

  _progress_bar.set_value(_current_length);
  _song_name.set_text(get_formatted_song_name(song));

  string orig_filename = song[SFI_filename].get_string();
  VariableDef vars;

  char index_str[128];
  sprintf(index_str, "%04d", _current_index);
  vars["index"] = string(index_str);
  vars["basename"] = get_basename_no_ext(orig_filename);
  vars["ext"] = get_filename_extension(orig_filename);
  vars["name"] = make_filename_friendly(get_formatted_song_name(song));
  vars["performer"] = 
    make_filename_friendly(get_formatted_artist_name(performer));
  vars["writer"] = 
    make_filename_friendly(get_formatted_artist_name(writer));
  vars["cdtitle"] = make_filename_friendly(get_formatted_cd_name(cd));
  vars["cdperformer"] =
    make_filename_friendly(get_formatted_artist_name(cdperformer));
  vars["cdwriter"] =
    make_filename_friendly(get_formatted_artist_name(cdwriter));

  char track_str[128];
  sprintf(track_str, "%02d", track_num);
  vars["track"] = string(track_str);
  
  string filename_to = expand_variables(_filename_pattern, vars);

  start_exporting(song, _root_dir + "/" + filename_to);
  return true;
}

void ExportingDialog::
start_exporting(const TableDataProxy &song, const string &filename_to) {
  string rel_filename = song[SFI_filename].get_string();
  Volume volume(_data, song[SFI_volume_id].get_foreign(TI_volumes));
  string filename_from;

  if (volume.is_valid()) {
    filename_from = volume.get_filename(rel_filename);
  }

  if (filename_from.empty()) {
    new InfoMessage("Unable to get source filename for\n" + song[SFI_name]);
    _canceled = true;
    export_failure();
    return;
  }

  string dirname = get_dirname(filename_to);
  mkdir_p(dirname);

  string export_command;

  if (_symlink) {
    // Symbolic links are so easy, we can do them directly.
    int result = symlink(filename_from.c_str(), filename_to.c_str());
    if (result != 0) {
      new InfoMessage("Unable to create symlink to " + filename_to +
		      ":\n" + strerror(errno));
      _canceled = true;
      export_failure();
      return;
    }

    // We could just loop back to export_success(), but this is a
    // single-threaded GUI application, and we have to yield control
    // from time to time.  We arrange to yield every second by
    // hanging a hook on idle() and then returning to main.
    time_t now = time(NULL);
    if (now > _last_return) {
      _last_return = now;
      connect_to_method(Gtk_Main::idle(), this, &back_from_idle);
    } else {
      export_success();
    }
    return;

  } else if (_convert_type < 0) {
    export_command = "cp " + protect_from_shell(filename_from) + " " + 
      protect_from_shell(filename_to);

  } else {
    // We have to build a real export command.
    string export_command_unexpanded;

    int from_type = song[SFI_file_type].get_int(-1);
    if (from_type < 0) {
      from_type = guess_file_type(rel_filename);
    }

    if (from_type < 0) {
      new InfoMessage("Unable to determine file type of " + filename_from);
      _canceled = true;
      export_failure();
      return;
    }

    // Get the export command.
    QueryResult qr = music_db->sql_command
      ("SELECT command FROM soundfile_convert"
       " WHERE from_type=" + DataCell(from_type) +
       " AND to_type=" + DataCell(_convert_type));
    if (qr.get_num_rows() > 0) {
      export_command_unexpanded = qr.get_cell(0, 0).get_string();
    }

    if (export_command_unexpanded.empty()) {
      if (_convert_type == from_type) {
	export_command_unexpanded = "cp $from $to";

      } else {
	TableDataProxy from_type_sf(TI_soundfile, from_type);
	TableDataProxy to_type_sf(TI_soundfile, _convert_type);
	new InfoMessage("Don't know how to convert files from\n" +
			from_type_sf[SFFI_description] + " to " +
			to_type_sf[SFFI_description] + ".  Please\n"
			"define a command for this conversion.\n");
	_canceled = true;
	export_failure();
	return;
      }
    }

    VariableDef vars;
    vars["from"] = protect_from_shell(filename_from);
    vars["to"] = protect_from_shell(filename_to);
    
    export_command =
      expand_variables(export_command_unexpanded, vars);
  }

  _shell.run(export_command);
}

string ExportingDialog::
make_filename_friendly(const string &orig_name) {
  string name = iso8859_to_ascii(orig_name);

  string filename;
  bool last_space = false;

  for (int i = 0; i < name.length(); i++) {
    if (isspace(name[i]) || name[i]=='/') {
      if (!last_space) {
	filename += '-';
	last_space = true;
      }

    } else {
      last_space = false;
      if (isprint(name[i])) {
	filename += name[i];
      }
    }
  }
  return filename;
}
