/* formDef.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "formDef.h"
#include "database.h"
#include "dataCell.h"

#include <algorithm>

typedef map<int, FormDef *> FormDefs;
static FormDefs form_defs;

FormDef::
FormDef() {
  _form_id = -1;
  _num_rows = 0;
  _num_columns = 0;
  _homogeneous = false;
  _x_size_pixels = 0;
  _y_size_pixels = 0;
  _scrollbar = false;
}

FormDef::
~FormDef() {
}

const FormDef *FormDef::
get_form(int form_id) {
  FormDefs::const_iterator edi;
  edi = form_defs.find(form_id);
  if (edi != form_defs.end()) {
    return (*edi).second;
  } else {
    FormDef *def = new FormDef;
    def->look_up(form_id);
    form_defs[form_id] = def;
    return def;
  }
}

const FormDef *FormDef::
reread_form(int form_id) {
  FormDefs::iterator edi;
  edi = form_defs.find(form_id);
  if (edi != form_defs.end()) {
    delete (*edi).second;
    FormDef *def = new FormDef;
    def->look_up(form_id);
    (*edi).second = def;
    return def;
  } else {
    FormDef *def = new FormDef;
    def->look_up(form_id);
    form_defs[form_id] = def;
    return def;
  }
}

void FormDef::
record_db() {
  if (_form_id >= 0) {
    // We are updating an existing form record.
    music_db->sql_command
      ("UPDATE forms SET "
       "name=" + DataCell(_name, !_name.empty()) + ", "
       "num_rows=" + DataCell(_num_rows) + ", "
       "num_columns=" + DataCell(_num_columns) + ", "
       "homogeneous=" + DataCellBool(_homogeneous) + ", "
       "x_size_pixels=" + DataCell(_x_size_pixels) + ", "
       "y_size_pixels=" + DataCell(_y_size_pixels) + ", "
       "scrollbar=" + DataCell(_scrollbar) + ", "
       "general_sql_command=" + 
       DataCell(_general_sql_command, !_general_sql_command.empty()) +
       " WHERE form_id=" + DataCell(_form_id));

    // Remove any old form widgets that might or might not be
    // there.
    music_db->sql_command
      ("DELETE FROM form_widgets WHERE form_id=" + DataCell(_form_id));
  } else {
    // We are inserting a new form record.
    _form_id = music_db->get_next_sequence("forms_form_id_seq");

    music_db->sql_command
      ("INSERT INTO forms (form_id, name, num_rows, num_columns,"
       " homogeneous, x_size_pixels, y_size_pixels, scrollbar,"
       " general_sql_command)"
       " VALUES (" + 
       DataCell(_form_id) + ", " +
       DataCell(_name, !_name.empty()) + ", " +
       DataCell(_num_rows) + ", " +
       DataCell(_num_columns) + ", " +
       DataCellBool(_homogeneous) + ", " +
       DataCell(_x_size_pixels) + ", " +
       DataCell(_y_size_pixels) + ", " +
       DataCellBool(_scrollbar) + ", " +
       DataCell(_general_sql_command, !_general_sql_command.empty()) +
       ")");
  }

  // Now insert new definitions for all of the widgets.
  Widgets::iterator wi;
  for (wi = _widgets.begin(); wi != _widgets.end(); ++wi) {
    (*wi).insert_db(_form_id);
  }
}

bool FormDef::
look_up(int form_id) {
  assert(form_id >= 0);
  _form_id = form_id;
  _name = "";
  _widgets.clear();

  QueryResult qr = music_db->sql_command
    ("SELECT name, num_rows, num_columns, homogeneous, x_size_pixels, "
     "y_size_pixels, scrollbar, general_sql_command"
     " FROM forms WHERE form_id=" + DataCell(_form_id));

  if (qr.get_num_rows() == 0) {
    return false;
  }

  _name = qr.get_cell(0, 0).get_string();
  _num_rows = qr.get_cell(0, 1).get_int();
  _num_columns = qr.get_cell(0, 2).get_int();
  _homogeneous = qr.get_cell(0, 3).get_bool();
  _x_size_pixels = qr.get_cell(0, 4).get_int();
  _y_size_pixels = qr.get_cell(0, 5).get_int();
  _scrollbar = qr.get_cell(0, 6).get_bool();
  _general_sql_command = qr.get_cell(0, 7).get_string();

  qr = music_db->sql_command
    ("SELECT widget_id, table_id, field_id, widget_type, type_assoc, "
     "place_left, place_right, place_top, place_bottom, xoptions, yoptions, "
     "xpadding, ypadding, uxsize, uysize, internal_name, label "
     "FROM form_widgets "
     "WHERE form_id=" + DataCell(_form_id));

  int num_rows = qr.get_num_rows();
  for (int r = 0; r < num_rows; r++) {
    FormWidget w;
    w._widget_id = qr.get_cell(r, 0).get_int();
    w._table_id = (TableID)qr.get_cell(r, 1).get_int(TI_no_table);
    w._field_id = qr.get_cell(r, 2).get_int(-1);
    w._widget_type = (FormWidget::Type)qr.get_cell(r, 3).get_int();
    w._type_assoc = (FormWidget::Type)qr.get_cell(r, 4).get_int(-1);
    w._place_left = qr.get_cell(r, 5).get_int();
    w._place_right = qr.get_cell(r, 6).get_int();
    w._place_top = qr.get_cell(r, 7).get_int();
    w._place_bottom = qr.get_cell(r, 8).get_int();
    w._xoptions = qr.get_cell(r, 9).get_int();
    w._yoptions = qr.get_cell(r, 10).get_int();
    w._xpadding = qr.get_cell(r, 11).get_int();
    w._ypadding = qr.get_cell(r, 12).get_int();
    w._uxsize = qr.get_cell(r, 13).get_int();
    w._uysize = qr.get_cell(r, 14).get_int();
    w._internal_name = qr.get_cell(r, 15).get_string();
    w._label = qr.get_cell(r, 16).get_string();
    
    _widgets.push_back(w);
  }

  return true;
}

int
get_related_form_id(int form_id, TableID table_id) {
  if (form_id < 0) {
    return -1;
  }

  QueryResult qr = music_db->sql_command
    ("SELECT next_form FROM form_relations"
     " WHERE form_id=" + DataCell(form_id) +
     " AND table_id=" + DataCell(table_id));

  int next_form = -1;

  if (qr.get_num_rows() != 0) {
    next_form = qr.get_cell(0, 0).get_int();
  }
  return next_form;
}
